#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <arpa/inet.h>
//#include <assert.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#define DEFAULT_LEN 	16
#define DEFAULT_PREV 	5
#define MAX_LINE_LEN	512

static char *version = "0.30";
static long port_number = 8080;

static char *mime_html = "Content-type: text/html\n\n";
static char *mime_xml = "Content-type: text/xml\n\n";

static char *http_200 = "HTTP/1.1 200 OK\n";
static char *http_400 = "HTTP/1.1 400 Bad Request\n";
static char *http_405 = "HTTP/1.1 405 Method Not Allowed\n";
static char *http_500 = "HTTP/1.1 500 Internal Server Error\n";

static void sigchld_action_handler(int signal_number) {
	int status;
	wait(&status);
}

void fatal_error(char *cause) {
	fprintf(stderr, "%s: error: (%s) %s\n", "empegweb", cause, strerror(errno));
	exit(1);
}

FILE *get_tagfile(int fid) {
	FILE *file;
	char path[] = "/empeg/fids0/XXXXXXXXXX";
	
	//get the fid tags file for reading (or with 1 to ensure tags fid)
	sprintf(path + 13, "%x", fid | 1);
	if((file = fopen(path, "ro")) == NULL) {
		path[11] ^= 1;
		if((file = fopen(path, "ro")) == NULL)
			fprintf(stderr, "%s: error: (%s) %s 0x%05x\n", "empegweb", "open", "could not open fid tags file for", fid);
	}
	return file;
}

int linetoxml_size(char *input) {
	char *ch, *position = strstr(input, "=");
	int size = ((position - input) * 2) + 7; //begin and end tags plus \t<></>\n
	input += (position - input);
	size += strlen(input);

	while((ch = input++)) {
		if(*ch == '\0') break;
		if(*ch == '<' || *ch == '>') size += 3; //&lt;&gt;
		if(*ch == '&') size += 4; //&amp;
		if(*ch == '"') size += 5; //&quot;
	}
	
	return size;
}

char *str_append(char *output, char *input) {
	memcpy(output, input, strlen(input));
	return output + strlen(input);
}

void linetoxml(char *output, char *input) {
	char *p, *tagname = input;

	while((p = input++))
		if(*p == '=') { input--; *input = '\0';	break; }

	output = str_append(output, "\t<");
	output = str_append(output, tagname);
	output = str_append(output, ">");
	input++;

	while((p = input++)) {
		if(*p == '<') 		output = str_append(output, "&lt;");
		else if(*p == '>') 	output = str_append(output, "&gt;");
		else if(*p == '&') 	output = str_append(output, "&amp;");
		else if(*p == '"') 	output = str_append(output, "&quot;");
		else if(*p == '\n') ;//skip newlines
		else { memcpy(output, p, 1); output++; }
		if(*p == '\0') { output--; break; }
	}
	
	output = str_append(output, "</");
	output = str_append(output, tagname);
	output = str_append(output, ">\n");
	*output = '\0'; 
}

static void handle_get(int connection_fd, char *page) {
	unsigned long fidTableIndex;
	int playListLen, runOrderLen, i;
	FILE *dynamicData, *tagfile;
	int track, prev, len, start, fid;
	char *pair, *input, *output;
	char buffer[MAX_LINE_LEN];
	
	//initialize our variables
	track = -1;
	prev = DEFAULT_PREV;
	len = DEFAULT_LEN;
	
	//parse our CGI args
	input = page + 2; //CGI input after the /?
	pair = strtok(input, "&"); //tokenize on &
	while(pair) {
		if(strncmp(pair, "track", 5) == 0) sscanf(pair, "track=%d", &track);
		if(strncmp(pair, "prev", 4) == 0) sscanf(pair, "prev=%d", &prev);
		if(strncmp(pair, "len", 3) == 0) sscanf(pair, "len=%d", &len);
		pair = strtok(NULL, "&");	
	}
	
	//if we got a track #, we can proceed
	if(track != -1) {
		if((dynamicData = fopen("/dev/hda3", "ro")) == NULL) {
			write(connection_fd, http_500, strlen(http_500));
			write(connection_fd, mime_html, strlen(mime_html));
			write(connection_fd, http_500, strlen(http_500));
		} else {
			write(connection_fd, http_200, strlen(http_200));
			write(connection_fd, mime_xml, strlen(mime_xml));
			//sprintf(buffer, "<pre>track: %d\nprev: %d\nlen: %d\n", track, prev, len);
			//write(connection_fd, buffer, strlen(buffer));
	
			fseek(dynamicData, 0x10, SEEK_SET);
			fread(&playListLen, sizeof(int), 1, dynamicData);
			fseek(dynamicData, 0x18, SEEK_SET);
			fread(&runOrderLen, sizeof(int), 1, dynamicData);

			sprintf(buffer, "<?xml version=\"1.0\"?>\n<playlist length=\"%d\" runorder=\"%d\">\n", playListLen, runOrderLen);
			write(connection_fd, buffer, strlen(buffer));
			
			//ensure we are not starting below 0
			start = track - prev < 0 ? 0 : track - prev;
			for(i = 0; (i < len) && (start + i < playListLen); i++) {
				//get our fid
				fseek(dynamicData, (0x200 + 8 * playListLen + (start + i) * 4), SEEK_SET);
				fread(&fidTableIndex, sizeof(unsigned long), 1, dynamicData);
				fseek(dynamicData, (0x200 + 8 * fidTableIndex), SEEK_SET);
				fread(&fid, sizeof(unsigned int), 1, dynamicData);
				
				if((tagfile = get_tagfile(fid)) == NULL) {
					fatal_error("fopen");
				} else {
					sprintf(buffer, "<track number=\"%d\" playing=\"%s\">\n", start + i, start + i == track ? "yes" : "no");
					write(connection_fd, buffer, strlen(buffer));
					//convert each line to XML
					while(fgets(buffer, MAX_LINE_LEN, tagfile) != NULL) {
						output = malloc(linetoxml_size(buffer));
						linetoxml(output, buffer);
						write(connection_fd, output, strlen(output));
						free(output);
					}
					sprintf(buffer, "</track>\n");
					write(connection_fd, buffer, strlen(buffer));
				}
				
				fclose(tagfile);
				//sprintf(buffer, "fid(%02d)(%04d): %s 0x%05x\n", i, start + i, start + i == track ? "+" : "-", fid);
				//write(connection_fd, buffer, strlen(buffer));
			}
			
			sprintf(buffer, "</playlist>\n");
			write(connection_fd, buffer, strlen(buffer));
			
			fclose(dynamicData);
		}
	} else {
		//invalid track parameter, or no track parameter provided
		write(connection_fd, http_400, strlen(http_400));
		write(connection_fd, mime_html, strlen(mime_html));
		write(connection_fd, http_400, strlen(http_400));
	}
}

static void handle_connection(int connection_fd) {
	char buffer[256];
	ssize_t bytes_read;

	bytes_read = read(connection_fd, buffer, sizeof(buffer) - 1);
	if(bytes_read > 0) {
		char method[sizeof(buffer)];
		char url[sizeof(buffer)];
		char protocol[sizeof(buffer)];

		buffer[bytes_read] = '\0';
		
		//first line of HTTP request
		sscanf(buffer, "%s %s %s", method, url, protocol);
		//discard all other input (acccept, client type, etc.)
		while(strstr(buffer, "\r\n\r\n") == NULL)
			bytes_read = read(connection_fd, buffer, sizeof(buffer));
		//if we had a problem reading, shut down
		if(bytes_read == -1) {
			close(connection_fd);
			return;
		}
		
		if(strcmp(protocol, "HTTP/1.0") && strcmp(protocol, "HTTP/1.1")) {
			//not 1.0 or 1.1, probably a bad request
			write(connection_fd, http_400, strlen(http_400));
			write(connection_fd, mime_html, strlen(mime_html));
			write(connection_fd, http_400, strlen(http_400));
		} else if(strcmp(method, "GET")) {
			//not a GET request, method not allowed
			write(connection_fd, http_405, strlen(http_405));
			write(connection_fd, mime_html, strlen(mime_html));
			write(connection_fd, http_405, strlen(http_405));
		} else {
			handle_get(connection_fd, url);
		}
	} else if (bytes_read == 0) {
		//client closed connection
	} else {
		//error on read	
		fatal_error("read");
	}
}

void server_run(uint16_t port) {
	struct in_addr local_address;
	struct sockaddr_in socket_address;
	struct sigaction sigchld_action;
	int server_socket;

	//signal handler for cleaning up children
	memset(&sigchld_action, 0, sizeof(sigchld_action));
	sigchld_action.sa_handler = &sigchld_action_handler;
	sigaction(SIGCHLD, &sigchld_action, NULL);

	//create a TCP socket
	server_socket = socket(PF_INET, SOCK_STREAM, 0);
	if(server_socket == -1)	fatal_error("socket");

	//socket structure where we listen for connections
	local_address.s_addr = INADDR_ANY;
	memset(&socket_address, 0, sizeof(socket_address));
	socket_address.sin_family = AF_INET;
	socket_address.sin_port = port;
	socket_address.sin_addr = local_address;

	//bind the socket
	if((bind(server_socket, &socket_address, sizeof(socket_address))) != 0) fatal_error("bind");
	//socket listen
	if((listen(server_socket, 10)) != 0) fatal_error("listen");

	while(1) {
		struct sockaddr_in remote_address;
		socklen_t address_length;
		int connection;
		pid_t child_pid;

		address_length = sizeof(remote_address);
		connection = accept(server_socket, &remote_address, &address_length);
		if(connection == -1) {
			if(errno == EINTR) {
				//call was interrupted by a signal
				continue;
			} else {
				//something bad went wrong
				fatal_error("accept");
			}
		}

		//fork a child to handle the connection
		child_pid = fork();
		if(child_pid == 0) {
			//child process
			close(server_socket);
			handle_connection(connection);
			close(connection);
			exit(0);
		} else if(child_pid > 0) {
			//parent process
			close(connection);
		}
	}
}

int main(int argc, const char *argv[]) {
	uint16_t port = (uint16_t)htons(port_number);
	
	fprintf(stderr, "XMLPlaylist version %s starting...\n", version);
	
	server_run(port);
	
	return 0;
}
